@@ -57,7 +57,7 @@ True
 __author__ = "Ludovico Magnocavallo <ludo\x40qix\x2eit>"
 __copyright__ = "Copyright 2007, Ludovico Magnocavallo"
 __license__ = "MIT"
-__version__ = "0.1"
+__version__ = "0.2"
 __revision__ = "$LastChangedRevision: 1401 $"[22:-2]
 __date__ = "$LastChangedDate: 2007-04-03 15:34:47 +0200 (Tue, 03 Apr 2007) $"[18:-2]
 
@@ -78,6 +78,7 @@ from collections import deque
 BUFSIZE = 65536
 
 class ChannelError(Exception): pass
+class ChannelDisconnect(Exception): pass
 
 class Channel(object):
     """Base class to handle two-way communications between processes. This type
@@ -115,7 +116,7 @@ class Channel(object):
             if self._wb:
                 self._quit = True
             else:
-                self.disconnect()
+                raise ChannelDisconnect, "quitting"
         self.handle_message(message)
         
     def handle_message(self, message):
@@ -147,12 +148,10 @@ class Channel(object):
                 self.logger.info("EAGAIN in channel %s, resource temporarily unavailable" % self)
                 return
             self.logger.warning("socket error %s in channel %s" % (e, self))
-            self.disconnect()
-            return
+            raise ChannelDisconnect, "socket error %s in channel %s" % (e, self)
         if not data:
-            self.logger.warning("no data in channel %s" % self.pid)
-            self.disconnect()
-            return
+            self.logger.debug("no data in channel %s" % self.pid)
+            raise ChannelDisconnect, "no data in channel %s" % self.pid
         while True:
             msglen = self._msglen
             data = "%s%s" % (self._rb, data)
@@ -186,17 +185,17 @@ class Channel(object):
             sent = self._sock.send(wb)
         except socket.error, e:
             self.logger.warning("socket error while writing in channel %s, socket error %s" % (self, e))
-            self.disconnect()
+            raise ChannelDisconnect, "socket error while writing in channel %s, socket error %s" % (self, e)
         if sent == 0:
             self.logger.warning("nothing written to socket in channel %s" % self)
-            self._disconnect()
+            raise ChannelDisconnect, "nothing written to socket in channel %s" % self
         if sent < len(wb):
             self._wb = wb[sent:]
         else:
             self._wb = ''
             self._poll.register(self._sock, select.POLLIN | select.POLLPRI)
             if self._quit:
-                self.disconnect()
+                raise ChannelDisconnect, "quitting"
         
     def disconnect(self, e=None):
         """Disconnect this channel, closing the socket and exiting the child
@@ -227,10 +226,10 @@ class MasterChannel(Channel):
     def _handle_message(self, msg):
         self._cb(self, msg)
         
-    def disconnect(self):
-        """We do not need to disconnect, as the master polling loop exits as soon
-        as all child processes have terminated.
+    def disconnect(self, e=None):
+        """Disconnect without exiting this process.
         """
+        self.logger.debug("disconnect for master channel %s called" % self)
         pass
     
 class SerialWorkerChannel(Channel):
@@ -253,16 +252,22 @@ class SerialWorkerChannel(Channel):
                 if fd != self._fileno:
                     self.logger.critical("event %s on unkown socket %s (our socket %s)" % (event, fd, self._fileno))
                     continue
-                if event & select.POLLIN or event | select.POLLPRI:
-                    self.read()
-                if event & select.POLLOUT and self._wb:
-                    self.write()
-                if event & select.POLLERR:
-                    self.logger.critical("error on channel %s" % self)
                 if event & select.POLLHUP:
+                    # remote socket has closed connection, we can only read
+                    # until the buffer is empty, POLLIN is set anyway so just
+                    # log the event
                     self.logger.info("HUP in poll for %s" % self)
+                try:
+                    if event & select.POLLIN or event | select.POLLPRI:
+                        self.read()
+                    if event & select.POLLOUT and self._wb:
+                        self.write()
+                except ChannelDisconnect, e:
+                    self.logger.debug("disconnection in loop for %s: %s" % (self, e))
                     self.disconnect()
                     break
+                if event & select.POLLERR:
+                    self.logger.critical("error on channel %s" % self)
                 elif event & select.POLLNVAL:
                     self.disconnect("Error in poll")
                     break
@@ -297,16 +302,22 @@ class WorkerChannel(Channel):
                 if fd != self._fileno:
                     self.logger.critical("event %s on unkown socket %s (our socket %s)" % (event, fd, self._fileno))
                     continue
-                if event & select.POLLIN or event | select.POLLPRI:
-                    self.read()
-                if event & select.POLLOUT and self._wb:
-                    self.write()
-                if event & select.POLLERR:
-                    self.logger.critical("error on channel %s" % self)
                 if event & select.POLLHUP:
+                    # remote socket has closed connection, we can only read
+                    # until the buffer is empty, POLLIN is set anyway so just
+                    # log the event
                     self.logger.info("HUP in poll for %s" % self)
+                try:
+                    if event & select.POLLIN or event | select.POLLPRI:
+                        self.read()
+                    if event & select.POLLOUT and self._wb:
+                        self.write()
+                except ChannelDisconnect, e:
+                    self.logger.debug("disconnection in loop for %s: %s" % (self, e))
                     self.disconnect()
                     break
+                if event & select.POLLERR:
+                    self.logger.critical("error on channel %s" % self)
                 elif event & select.POLLNVAL:
                     self.disconnect("Error in poll")
                     break
@@ -417,15 +428,21 @@ class ChattyParallel(object):
                 channel = self._channels.get(fd)
                 if not channel:
                     self.logger.critical("event %s on unkown channel with socket %s" % (event, fd))
-                if event & select.POLLIN or event | select.POLLPRI:
-                    channel.read()
-                if event & select.POLLOUT:
-                    channel.write()
+                if event & select.POLLHUP:
+                    # remote socket has closed connection, we can only read
+                    # until the buffer is empty, POLLIN is set anyway so just
+                    # log the event
+                    self.logger.info("HUP in poll for %s" % self)
+                try:
+                    if event & select.POLLIN or event | select.POLLPRI:
+                        channel.read()
+                    if event & select.POLLOUT:
+                        channel.write()
+                except ChannelDisconnect, e:
+                    self.logger.debug("disconnection in loop for %s: %s" % (self, e))
+                    self._disconnect(fd)
                 if event & select.POLLERR:
                     self.logger.critical("error on channel %s" % channel)
-                if event & select.POLLHUP:
-                    self.logger.info("hup on channel %s" % channel)
-                    self._disconnect(fd)
                 elif event & select.POLLNVAL:
                     self.logger.critical("disconnect on channel %s" % channel)
                     self._disconnect(fd)
@@ -458,4 +475,4 @@ class ChattyParallel(object):
 if __name__ == '__main__':
     import doctest
     doctest.testmod()
-    
\ No newline at end of file
+    
